# $Id: Geocoder.pm,v 1.4 2005/11/06 17:49:27 asc Exp $
use strict;

package Yahoo::Geocoder;
use base qw (LWP::UserAgent);

$Yahoo::Geocoder::VERSION = '1.0';

=head1 NAME

Yahoo::Geocoder - Simple OO wrapper for Yahoo geocoder REST API

=head1 SYNOPSIS

  use Yahoo::Geocoder;
  use Error qw(:try);

  my $geo   = Yahoo::Geocoder->new({appid=>"yer_appid_here"});
  my $xpath = undef;

  try { 
      $xpath = $geo->geocode({location=>"Montreal, QC"});
  }

  catch Yahoo::Geocoder::ConfigError with {
      my $e = shift;
      die "There's something wrong with your request. $e";
  }

  catch Yahoo::Geocoder::HTTPError with {
      my $e = shift;
      die "There was a problem talking to Yahoo! $e";
  }

  catch Yahoo::Geocoder::XMLParseError with {
      my $e = shift;
      die "There was a problem with the data sent back by Yahoo! $e";
  }

  otherwise {
      my $e = shift;
      die "OMGWTF?!?! $e";
  };
      
  foreach my $place ($xpath->findnodes("/ResultSet/Result")) {
      print $place->findvalue("Latitude");
  }

  # And for people who prefer good
  # old-fashion hash references - for 
  # the sake of brevity, we'll assume
  # the request throws no errors

  my @results = $geo->geocode_as_list({location=>"Springfield"});

  foreach my $place (@results) {
      print $place->{Latitude};
  }

=head1 DESCRIPTION

Simple OO wrapper for Yahoo geocoder REST API.

=cut
  
use HTTP::Request;
use XML::XPath;
use Readonly;
use Error qw (:try);

Readonly::Scalar my $GEOCODER_SERVICE_URL         => "http://api.local.yahoo.com/MapsService/V1/geocode";
Readonly::Array  my @GEOCODER_REQUEST_PARAMS      => ("street","city","state","city","location","appid");
Readonly::Array  my @GEOCODER_RESPONSE_FIELDS     => ("Latitude","Longitude","Address","City","State","Zip","Country");
Readonly::Array  my @GEOCODER_RESPONSE_ATTRIBUTES => ("precision","warning");

=head1 PACKAGE METHODS

=cut

=head2 __PACKAGE__->new(\%args)

Valid arguments are :

=over 4

=item * B<appid>

String.

Your Yahoo! application ID.

=back

Returns a I<Yahoo::Geocoder> object, which subclasses I<LWP::UserAgent>

=cut

sub new {
  my $pkg  = shift;
  my $args = shift;

  my $self = $pkg->SUPER::new();
  $self->{appid} = $args->{appid};

  return bless $self, $pkg;
}

=head2 __PACKAGE__->service_url()

Returns a string.

=cut

sub service_url {
    return $GEOCODER_SERVICE_URL;
}

=head2 __PACKAGE__->request_params()

Returns a list.

=cut

sub request_params {
    return @GEOCODER_REQUEST_PARAMS;
}

=head2 __PACKAGE__->response_fields()

Returns a list.

=cut

sub response_fields {
    return @GEOCODER_RESPONSE_FIELDS;
}

=head2 __PACKAGE__->response_attributes()

Returns a list.

=cut

sub response_attributes {
    return @GEOCODER_RESPONSE_ATTRIBUTES;
}

=head1 OBJECT METHODS

=cut

=head2 $obj->geocode(\%args)

Valid arguments are :

=over 4

=item * B<address>

String.

=item * B<city>

String.

=item * B<state>

String.

=item * B<zip>

String.

=item * B<appid>

String.

=back

At least one argument, not including I<appid> is required. If <appid> is present,
it will override any application ID passed to the object's constructor.

Returns a I<XML::XPath> object on success. Throws an I<Error::Simple> 
derived exception if there was a problem.

=cut

sub geocode {
  my $self = shift;
  my $args = shift;

  my $appid = $self->{appid};

  my @params = ();

  foreach my $field ($self->request_params()) {
    if (! $args->{$field}) {
      next;
    }

    if ($field eq "appid") {
      $appid = $args->{$field};
      next;
    }

    push @params, sprintf("%s=%s",$field,$args->{$field});
  }

  if (! $appid) {
      throw Yahoo::Geocoder::ConfigError("No appid defined");
  }

  if (! @params) {
      throw Yahoo::Geocoder::ConfigError("No parameters defined");
  }

  my $url = sprintf("%s?appid=%s&%s",$self->service_url(),$appid,join("&",@params));

  try {
      $self->api_call($url);
  }

  otherwise {
      shift->throw();
  };

}

=head2 $obj->geocode_as_list(\%args)

This is a clone of the I<geocode> method whose only difference
is it's return value.

Returns a list of hash references on success. Throws an I<Error::Simple> 
derived exception if there was a problem.

=cut

sub geocode_as_list {
  my $self = shift;
  my $args = shift;
  my $xml  = undef;

  try {
      $xml = $self->geocode($args);
  }

  otherwise {
      shift->throw();
  };

  return map {
      $self->node_to_hashref($_);
  } $xml->findnodes("/ResultSet/Result");
}

=head2 $api_call($url)

Returns a I<XML::XPath> object on success. Throws an I<Error::Simple>
derived exception if there was a problem.

=cut

sub api_call {
    my $self = shift;
    my $url  = shift;

    try {
	my $res = $self->do_request($url);
	return $self->parse_response($res);
    }

    otherwise {
	shift->throw();
    };
}

=head2 $obj->do_request($url)

Returns a I<HTTP::Response> object on success. Throws an I<Error::Simple> derived
exception if there was a problem.

=cut

sub do_request {
    my $self = shift;
    my $url  = shift;

    my $req = HTTP::Request->new(GET=>$url);
    my $res = $self->request($req);

    if ($res->code() != 200) {
	throw Yahoo::Geocoder::HTTPError($res->code());
    }

    return $res;
}

=head2 $obj->parse_response(HTTP::Response)

Returns a I<XML::XPath> object on success. Throws an I<Error::Simple>
derived exception if there was a problem.

=cut

sub parse_response {
    my $self     = shift;
    my $response = shift;

    try {
	my $xml = XML::XPath->new(xml=>$response->content());
	return $xml;
    }
  
    otherwise {
	throw Yahoo::Geocoder::XMLParseError($@);
    };
}

=head2 $obj->node_to_hashref(XML::XPath::Node)

Returns a hash reference.

=cut

sub node_to_hashref {
    my $self = shift;
    my $result = shift;

    my %data = map {
	$_ => (($result->findvalue($_)) ? $result->findvalue($_)->value() : undef);
    } ($self->response_fields());
    
    map {
	$data{$_} = $result->getAttribute($_);
    } ($self->response_attributes());
    
    return \%data;
}

package Yahoo::Geocoder::ConfigError;
use base qw (Error::Simple);

package Yahoo::Geocoder::HTTPError;
use base qw (Error::Simple);

package Yahoo::Geocoder::XMLParseError;
use base qw (Error::Simple);

=head1 VERSION

1.0

=head1 DATE

$Date: 2005/11/06 17:49:27 $

=head1 AUTHOR

Aaron Straup Cope E<lt>ascope@cpan.orgE<gt>

=head1 SEE ALSO

http://developer.yahoo.net/maps/rest/V1/geocode.html

=head1 LICENSE

Copyright (c) 2005 Aaron Straup Cope. All Rights Reserved.

This is free software. You may redistribute it and/or modify
it under the same terms as Perl itself. 

=cut

return 1;
